@@ -1,6 +1,6 @@
 /* test/test_bgzf.c -- bgzf unit tests
 
-   Copyright (C) 2017, 2019 Genome Research Ltd
+   Copyright (C) 2017, 2019, 2022-2024 Genome Research Ltd
 
    Author: Robert Davies <rmd@sanger.ac.uk>
 
@@ -34,9 +34,11 @@ DEALINGS IN THE SOFTWARE.
 #include <inttypes.h>
 #include <fcntl.h>
 #include <unistd.h>
-#include "htslib/bgzf.h"
-#include "htslib/hfile.h"
-#include "hfile_internal.h"
+
+#include "../htslib/bgzf.h"
+#include "../htslib/hfile.h"
+#include "../htslib/hts_log.h"
+#include "../hfile_internal.h"
 
 const char *bgzf_suffix = ".gz";
 const char *idx_suffix  = ".gzi";
@@ -158,20 +160,26 @@ static BGZF * try_bgzf_hopen(const char *name, const char *mode,
     return bgz;
 }
 
-static int try_bgzf_close(BGZF **bgz, const char *name, const char *func) {
+static int try_bgzf_close(BGZF **bgz, const char *name, const char *func, int expected_fail) {
     BGZF *to_close = *bgz;
     *bgz = NULL;
     if (bgzf_close(to_close) != 0) {
-        fprintf(stderr, "%s : bgzf_close failed on %s : %s\n",
-                func, name, strerror(errno));
+        if (!expected_fail)
+            fprintf(stderr, "%s : bgzf_close failed on %s%s%s\n",
+                    func, name,
+                    errno ? " : " : "",
+                    errno ? strerror(errno) : "");
         return -1;
+    } else if (expected_fail) {
+        fprintf(stderr, "%s : bgzf_close worked on %s, but expected failure\n",
+                func, name);
     }
     return 0;
 }
 
 static ssize_t try_bgzf_read(BGZF *fp, void *data, size_t length,
                              const char *name, const char *func) {
-    ssize_t got = bgzf_read(fp, data, length);
+    ssize_t got = bgzf_read_small(fp, data, length);
     if (got < 0) {
         fprintf(stderr, "%s : Error from bgzf_read %s : %s\n",
                 func, name, strerror(errno));
@@ -181,7 +189,7 @@ static ssize_t try_bgzf_read(BGZF *fp, void *data, size_t length,
 
 static ssize_t try_bgzf_write(BGZF *fp, const void *data, size_t length,
                               const char *name, const char *func) {
-    ssize_t put = bgzf_write(fp, data, length);
+    ssize_t put = bgzf_write_small(fp, data, length);
     if (put < (ssize_t) length) {
         fprintf(stderr, "%s : %s %s : %s\n",
                 func, put < 0 ? "Error writing to" : "Short write on",
@@ -378,7 +386,7 @@ static int setup(const char *src, Files *f) {
         perror(__func__);
         goto fail;
     }
-    for (i = 0; i < max; i++) snprintf(text + i*8, text_sz - i*8, "%07d\n", i);
+    for (i = 0; i < max; i++) snprintf(text + i*8, text_sz - i*8, "%07u\n", i);
     f->text = (unsigned char *) text;
     f->ltext = text_sz - 1;
 
@@ -397,6 +405,7 @@ static int test_read(Files *f) {
     ssize_t bg_got, f_got;
     unsigned char bg_buf[BUFSZ], f_buf[BUFSZ];
 
+    errno = 0;
     bgz = try_bgzf_open(f->src_bgzf, "r", __func__);
     if (!bgz) return -1;
 
@@ -413,7 +422,7 @@ static int test_read(Files *f) {
         }
     } while (bg_got > 0 && f_got > 0);
 
-    if (try_bgzf_close(&bgz, f->src_bgzf, __func__) != 0) return -1;
+    if (try_bgzf_close(&bgz, f->src_bgzf, __func__, 0) != 0) return -1;
     if (try_fseek_start(f->f_plain, f->src_plain, __func__) != 0) return -1;
 
     return 0;
@@ -448,7 +457,7 @@ static int test_write_read(Files *f, const char *mode, Open_method method,
     bg_put = try_bgzf_write(bgz, f->text, f->ltext, f->tmp_bgzf, __func__);
     if (bg_put < 0) goto fail;
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     switch (method) {
     case USE_BGZF_DOPEN:
@@ -490,7 +499,7 @@ static int test_write_read(Files *f, const char *mode, Open_method method,
         goto fail;
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     return 0;
 
@@ -520,7 +529,7 @@ static int test_embed_eof(Files *f, const char *mode, int nthreads) {
     bg_put = try_bgzf_write(bgz, f->text, half, f->tmp_bgzf, __func__);
     if (bg_put < 0) goto fail;
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
 
     // Write second half.  Append mode, so an EOF block should be in the
@@ -534,7 +543,7 @@ static int test_embed_eof(Files *f, const char *mode, int nthreads) {
                             __func__);
     if (bg_put < 0) goto fail;
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     // Try reading
     pos = 0;
@@ -563,7 +572,7 @@ static int test_embed_eof(Files *f, const char *mode, int nthreads) {
         goto fail;
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     return 0;
 
@@ -600,7 +609,7 @@ static int test_index_load_dump(Files *f) {
     } while (got_src > 0 && got_dest > 0);
     if (try_fclose(&fdest, f->tmp_idx, __func__) != 0) goto fail;
 
-    if (try_bgzf_close(&bgz, f->src_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->src_bgzf, __func__, 0) != 0) goto fail;
 
     return 0;
 
@@ -623,7 +632,7 @@ static int test_check_EOF(char *name, int expected) {
         return -1;
     }
 
-    return try_bgzf_close(&bgz, name, __func__);
+    return try_bgzf_close(&bgz, name, __func__, 0);
 }
 
 static int test_index_useek_getc(Files *f, const char *mode,
@@ -650,7 +659,7 @@ static int test_index_useek_getc(Files *f, const char *mode,
         }
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     bgz = try_bgzf_open(f->tmp_bgzf, "r", __func__);
     if (!bgz) goto fail;
@@ -709,7 +718,7 @@ static int test_index_useek_getc(Files *f, const char *mode,
         }
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     return 0;
 
@@ -740,7 +749,7 @@ static int test_tell_seek_getc(Files *f, const char *mode,
         if (bg_put < 0) goto fail;
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     bgz = try_bgzf_open(f->tmp_bgzf, "r", __func__);
     if (!bgz) goto fail;
@@ -810,7 +819,7 @@ static int test_tell_seek_getc(Files *f, const char *mode,
         }
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     return 0;
 
@@ -840,7 +849,7 @@ static int test_tell_read(Files *f, const char *mode) {
         if (bg_put < 0) goto fail;
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     bgz = try_bgzf_open(f->tmp_bgzf, "r", __func__);
     if (!bgz) goto fail;
@@ -858,7 +867,7 @@ static int test_tell_read(Files *f, const char *mode) {
         }
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
     free(bg_buf);
     return 0;
 
@@ -869,6 +878,49 @@ static int test_tell_read(Files *f, const char *mode) {
     return -1;
 }
 
+static int test_useek_read_small(Files *f, const char *mode) {
+
+    BGZF* bgz = NULL;
+    char bg_buf[99];
+
+    bgz = try_bgzf_open(f->tmp_bgzf, mode, __func__);
+    if (!bgz) goto fail;
+
+
+    if (try_bgzf_write(bgz, "#>Hello, World!\n", 16,
+                       f->tmp_bgzf, __func__) != 16)
+        goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
+
+    bgz = try_bgzf_open(f->tmp_bgzf, "r", __func__);
+    if (!bgz) goto fail;
+
+    if (try_bgzf_getc(bgz, 0, '#', f->tmp_bgzf, __func__) < 0 ||
+        try_bgzf_getc(bgz, 1, '>', f->tmp_bgzf, __func__) < 0)
+        goto fail;
+
+    if (try_bgzf_read(bgz, bg_buf, 5, f->tmp_bgzf, __func__) != 5)
+        goto fail;
+    if (memcmp(bg_buf, "Hello", 5) != 0)
+        goto fail;
+
+    if (try_bgzf_useek(bgz, 9, SEEK_SET, f->tmp_bgzf, __func__) < 0)
+        goto fail;
+
+    if (try_bgzf_read(bgz, bg_buf, 5, f->tmp_bgzf, __func__) != 5)
+        goto fail;
+    if (memcmp(bg_buf, "World", 5) != 0)
+        goto fail;
+
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
+    return 0;
+
+ fail:
+    fprintf(stderr, "%s: failed\n", __func__);
+    if (bgz) bgzf_close(bgz);
+    return -1;
+}
+
 static int test_bgzf_getline(Files *f, const char *mode, int nthreads) {
     BGZF* bgz = NULL;
     ssize_t bg_put;
@@ -884,11 +936,13 @@ static int test_bgzf_getline(Files *f, const char *mode, int nthreads) {
     bg_put = try_bgzf_write(bgz, f->text, f->ltext, f->tmp_bgzf, __func__);
     if (bg_put < 0) goto fail;
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
 
     bgz = try_bgzf_open(f->tmp_bgzf, "r", __func__);
     if (!bgz) goto fail;
 
+    if (nthreads > 0 && try_bgzf_mt(bgz, nthreads, __func__) != 0) goto fail;
+
     for (pos = 0; pos < f->ltext; ) {
         const char *end = strchr(text + pos, '\n');
         size_t l = end ? end - (text + pos) : f->ltext - pos;
@@ -908,12 +962,13 @@ static int test_bgzf_getline(Files *f, const char *mode, int nthreads) {
                     "Got      : %.*s\n",
                     __func__, f->tmp_bgzf, (int) l, (char *) f->text + pos,
                     (int) str.l, str.s);
+            goto fail;
         }
 
         pos += l + 1;
     }
 
-    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__) != 0) goto fail;
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
     free(ks_release(&str));
     return 0;
 
@@ -923,6 +978,98 @@ static int test_bgzf_getline(Files *f, const char *mode, int nthreads) {
     return -1;
 }
 
+static int test_bgzf_getline_on_truncated_file(Files *f, const char *mode, int nthreads) {
+    BGZF* bgz = NULL;
+    ssize_t bg_put;
+    size_t pos;
+    kstring_t str = { 0, 0, NULL };
+    const char *text = (const char *) f->text;
+
+    // Turn off bgzf errors as they're expected.
+    enum htsLogLevel lvl = hts_get_log_level();
+    hts_set_log_level(HTS_LOG_OFF);
+
+    bgz = try_bgzf_open(f->tmp_bgzf, mode, __func__);
+    if (!bgz) goto fail;
+
+    if (nthreads > 0 && try_bgzf_mt(bgz, nthreads, __func__) != 0) goto fail;
+
+    const char *text_line2 = strchr(text, '\n') + 1;
+    bg_put = try_bgzf_write(bgz, text, text_line2 - text, f->tmp_bgzf, __func__);
+    if (bg_put < 0) goto fail;
+    if (bgzf_flush(bgz) < 0) goto fail;
+    int64_t block2_start = bgz->block_address;
+
+    const char *text_line3 = strchr(text_line2, '\n') + 1;
+    bg_put = try_bgzf_write(bgz, text_line2, text_line3 - text_line2, f->tmp_bgzf, __func__);
+    if (bg_put < 0) goto fail;
+    if (bgzf_flush(bgz) < 0) goto fail;
+    int64_t block3_start = bgz->block_address;
+
+    if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 0) != 0) goto fail;
+
+    int64_t newsize;
+    for(newsize = block3_start - 1; newsize > block2_start; newsize--) {
+        //fprintf(stderr, "test_bgzf_getline_on_truncated_file : size truncated to %" PRId64 " with threads %d\n", newsize, nthreads);
+
+        if (truncate(f->tmp_bgzf, newsize) != 0) goto fail;
+
+        bgz = try_bgzf_open(f->tmp_bgzf, "r", __func__);
+        if (!bgz) goto fail;
+
+        if (nthreads > 0 && try_bgzf_mt(bgz, nthreads, __func__) != 0) goto fail;
+
+        for (pos = 0; pos < f->ltext; ) {
+            const char *end = strchr(text + pos, '\n');
+            size_t l = end ? end - (text + pos) : f->ltext - pos;
+
+            int res = bgzf_getline(bgz, '\n', &str);
+            if (res < -1) {
+                // ok, we expect error from truncated file
+                break;
+            } else if (res == -1) {
+                // truncated file should never return EOF since we do not truncate at block boundary
+                fprintf(stderr, "%s : %s from bgzf_getline on %s\n",
+                        __func__, "Unexpected EOF",
+                        f->tmp_bgzf);
+                goto fail;
+            }
+
+            if (str.l != l || memcmp(text + pos, str.s, l) != 0) {
+                fprintf(stderr,
+                        "%s : Unexpected data from bgzf_getline on %s\n"
+                        "Expected : %.*s\n"
+                        "Got      : %.*s\n",
+                        __func__, f->tmp_bgzf, (int) l, (char *) f->text + pos,
+                        (int) str.l, str.s);
+                goto fail;
+            }
+            pos += l + 1;
+        }
+
+        // verify error is persistent
+        int k;
+        for(k = 0; k < 3; k++) {
+            int res = bgzf_getline(bgz, '\n', &str);
+            if (res > -2) {
+                fprintf(stderr, "%s : unexpected bgzf_getline result %d\n", __func__, res);
+                goto fail;
+            }
+        }
+        // closing a stream with error returns error
+        if (try_bgzf_close(&bgz, f->tmp_bgzf, __func__, 1) == 0) goto fail;
+    }
+    free(ks_release(&str));
+    hts_set_log_level(lvl);
+    return 0;
+
+ fail:
+    hts_set_log_level(lvl);
+    if (bgz) bgzf_close(bgz);
+    free(ks_release(&str));
+    return -1;
+}
+
 int main(int argc, char **argv) {
     Files f = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 0 };
     int retval = EXIT_FAILURE;
@@ -994,11 +1141,19 @@ int main(int argc, char **argv) {
     if (test_tell_read(&f, "w") != 0) goto out;
     if (test_tell_read(&f, "wu") != 0) goto out;
 
+    // bgzf_useek and bgzf_read_small
+    if (test_useek_read_small(&f, "w") != 0) goto out;
+    if (test_useek_read_small(&f, "wu") != 0) goto out;
+
     // getline
     if (test_bgzf_getline(&f, "w", 0) != 0) goto out;
     if (test_bgzf_getline(&f, "w", 1) != 0) goto out;
     if (test_bgzf_getline(&f, "w", 2) != 0) goto out;
 
+    if (test_bgzf_getline_on_truncated_file(&f, "w", 0) != 0) goto out;
+    if (test_bgzf_getline_on_truncated_file(&f, "w", 1) != 0) goto out;
+    if (test_bgzf_getline_on_truncated_file(&f, "w", 2) != 0) goto out;
+
     retval = EXIT_SUCCESS;
 
  out:
